module snag.core.core;

import snag.config;
import std.format;
import std.datetime : Clock;
import std.stdio;
import std.array;
import std.process;
import std.algorithm;
import std.string;
import std.regex;

import snag.core.exception;

class Snag {
	private string[] _baseCommand;
	private SnagConfig _config;
	private string _date;

	private bool isValidHash(string hash) {
		auto hashPattern = ctRegex!r"^[a-fA-F0-9]{7}$";
		return !matchFirst(hash, hashPattern).empty;
	}

	this(SnagConfig config) {
		_baseCommand = format(
			"git --git-dir=%s --work-tree=%s",
			config.git, config.project
		).split();
		_config = config;

		auto currentTime = Clock.currTime();

		_date = format("%02d%03d%02d%02d%02d",
						currentTime.year % 100,
						currentTime.dayOfYear,
						currentTime.hour,
						currentTime.minute,
						currentTime.second);
	}

	private auto git(string[] command, string message, string separator = ":\n") {
		auto result = execute(_baseCommand ~ command);
		if (result.status)
			throw new SnagException(
				message ~ separator ~ result.output
			);
		return result;
	}

	void initialize() {
		git(
			["init"],
			"A Git repository initialization error occurred"
		);
		git(
			["config", "user.email", _config.email],
			"A Git repository initialization error occurred"
		);
		git(
			["config", "user.name", _config.user],
			"A Git repository initialization error occurred"
		);
		writeln(
			"The Git repository has been initialized successfully: ",
			_config.git
		);
	}

	void status() {
		auto result = git(
			["status", "--porcelain"],
			"An error occurred while checking the file tracking status"
		);

		if (!result.output.length) {
			writeln("The current state of the files is up to date as of the latest snapshot");
			return;
		}

		writeln("The following list of files requires backup:");
		result.output.split('\n').filter!(e => !e.strip.empty).each!((e) {
			writefln("\t/%s", e.strip.split[1]);
		});
	}

	void create() {
		auto result = git(
			["status", "--porcelain"],
			"An error occurred while checking the file tracking status"
		);

		// Если нечего коммитить, то выходим
		if (!result.output.length)
			throw new SnagException(
				"Current file state doesn't need to be archived again"
			);

		result = execute(_baseCommand ~ ["rev-parse", "--short", "HEAD"]);
		if (result.status == 128) {
			// Если это самый первый коммит после инициализации репозитория
			git(["add", "."], "Failed to prepare files for archiving");
			git(["commit", "-m", "test"], "Failed to create a backup");
			writeln("Backup was created successfully");
			return;
		} else if (result.status != 0)
			throw new SnagException(
				"Failed to retrieve current snapshot information:\n"
				~ result.output
			);

		// Текущий коммит, который был изменен
		string currentSnapshot = result.output.strip('\n');

		// Если текущий измененный коммит является последним в ветке - то продолжить коммиты в этой ветке
		string currentBranch = git(
			["for-each-ref", "--contains", currentSnapshot, "--format='%(refname:short)'"],
			"Error while getting the current branch"
		).output.split('\n')[0].strip('\'');

		// Получение списка коммитов между текущим и веткой
		result = git(
			["log", "--oneline", "HEAD.." ~ currentBranch],
			"Failed to get the commit list between HEAD and " ~ currentBranch
		);

		string newSnapshot;

		// Если список существует
		if (result.output.length) {
			// Если коммит не является последним, то необходимо ответвление
			string newBranch = "%s-%s".format(_date, currentSnapshot);
			git(
				["checkout", "-b", newBranch, currentSnapshot ],
				"Failed to create a branch from the current state"
			);
			git(
				["add", "."],
				"Failed to prepare files for archiving"
			);
			git(
				["commit", "-m", "test"],
				"Failed to create a backup"
			);
			newSnapshot = git(
				["rev-parse", "--short", "HEAD"],
				"Failed to retrieve current snapshot information"
			).output.strip('\n');
		} else {
			// Если коммит является посленим в ветке
			git(
				["add", "."],
				"Failed to prepare files for archiving"
			);
			git(
				["commit", "-m", "test"],
				"Failed to create a backup"
			);
			newSnapshot = git(
				["rev-parse", "--short", "HEAD"],
				"Failed to retrieve current snapshot information"
			).output.strip('\n');
			git(
				["checkout", currentBranch],
				"Failed to perform intermediate state switching"
			);
			git(
				["merge", newSnapshot],
				"Issue with including the commit into the branch " ~ currentBranch
			);
		}
		writeln("Backup was created successfully: ", newSnapshot);
	}

	void list(bool comment, bool user, bool email) {
		string currentSnapshot = git(
			["rev-parse", "--short", "HEAD"],
			"Failed to retrieve current snapshot information"
		).output.strip('\n');

		string format = "format:%h\t%ad";

		comment && (format ~= "\t%s");
		user && (format ~= "\t%an");
		email && (format ~= "\t%ae");

		git(
			[
				"log",
				"--all",
				"--date=format:%Y.%m.%d %H:%M",
				"--pretty=" ~ format
			],
			"Failed to retrieve the list of snapshots"
		).output.split('\n').map!(line => line.split('\t')).array.each!(e =>
        	writefln("%s\t%s",
				currentSnapshot == e[0] ? "    >" : "",
				e.join("\t")
			)
		);
	}

	void restore(string hash) {
		if (!isValidHash(hash))
			throw new SnagException(
				"Invalid snapshot hash provided"
			);

		auto result = git(
			["status", "--porcelain"],
			"An error occurred while checking the file tracking status"
		);

		if (result.output.length)
			git(
				["restore", "."],
				"Failed to reset file changes state"
			);

		git(
			["rev-parse", hash],
			"This snapshot is not available in the archive"
		);
		git(
			["checkout", hash],
			"Failed to restore the snapshot state " ~ hash
		);
		writeln("Backup was restored successfully: ", hash);
	}
}
